# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional, Dict

from .HeavyException import HeavyException
from .HeavyIrObject import HeavyIrObject
from .HeavyLangObject import HeavyLangObject
from .HeavyGraph import HeavyGraph


class HLangVario(HeavyLangObject):

    def __init__(
        self,
        obj_type: str,
        args: Dict,
        graph: 'HeavyGraph',
        annotations: Optional[Dict] = None
    ) -> None:
        assert obj_type == "vario"
        super().__init__(obj_type, args, graph, annotations=annotations)

    def reduce(self) -> Optional[tuple]:
        if self.graph is not None and self.name is not None:
            var_obj = self.graph.resolve_object_for_name(self.name, "var")
            if var_obj is None:
                raise HeavyException(
                    f"No corresponding \"var\" object found for object {self} in file {self.graph.file}"
                )

            if self.has_inlet_connection_format("f") and self.has_outlet_connection_format("_"):
                x = HeavyIrObject("__varwrite~f", {"var_id": var_obj.id})
                return {x}, self.get_connection_move_list(x)

            elif self.has_inlet_connection_format("i") and self.has_outlet_connection_format("_"):
                x = HeavyIrObject("__varwrite~i", {"var_id": var_obj.id})
                return {x}, self.get_connection_move_list(x)

            elif self.has_inlet_connection_format("_") and self.has_outlet_connection_format("f"):
                x = HeavyIrObject("__varread~f", {"var_id": var_obj.id})
                return {x}, self.get_connection_move_list(x)

            elif self.has_inlet_connection_format("_") and self.has_outlet_connection_format("i"):
                x = HeavyIrObject("__varread~i", {"var_id": var_obj.id})
                return {x}, self.get_connection_move_list(x)

            else:
                raise Exception()

        return None



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional, List, Dict

from .NotificationEnum import NotificationEnum
from .PdObject import PdObject
from .PdRaw import parse_pd_raw_args, PdRawException


class PdSendObject(PdObject):
    def __init__(
        self,
        obj_type: str,
        obj_args: Optional[List] = None,
        pos_x: int = 0,
        pos_y: int = 0
    ) -> None:
        assert obj_type in {"s", "send", "s~", "send~", "throw~"}
        PdObject.__init__(self, obj_type, obj_args, pos_x, pos_y)

        self.__send_name = ""
        self.__extern_type = None
        self.__attributes: Dict = {}

        try:
            # send objects don't necessarily need to have a name
            self.__send_name = self.obj_args[0]

            # only extern control rate sends
            if obj_type in {"s", "send"}:
                if self.obj_args[1] == "@hv_param":
                    self.__extern_type = "param"
                elif self.obj_args[1] == "@hv_event":
                    self.__extern_type = "event"
        except Exception:
            pass

        if self.__extern_type == "param":
            try:
                self.__attributes = {
                    "min": 0.0,
                    "max": 1.0,
                    "default": 0.5,
                    "type": "float"
                }
                self.__attributes["min"] = float(self.obj_args[2])
                self.__attributes["max"] = float(self.obj_args[3])
                self.__attributes["default"] = float(self.obj_args[4])
                self.__attributes["type"] = str(self.obj_args[5])
            except ValueError:
                self.add_warning(
                    f"Minimum, maximum, and default values for Parameter {self.__send_name}  must be numbers.")
            except Exception:
                pass

            if not (self.__attributes["min"] <= self.__attributes["default"]):
                self.add_error("Default parameter value is less than the minimum. "
                               "Send will not be exported: {0:g} < {1:g}".format(
                                   self.__attributes["default"],
                                   self.__attributes["min"]))
                self.__extern_type = None
            if not (self.__attributes["default"] <= self.__attributes["max"]):
                self.add_error("Default parameter value is greater than the maximum. "
                               "Send will not be exported: {0:g} > {1:g}".format(
                                   self.__attributes["default"],
                                   self.__attributes["max"]))
                self.__extern_type = None

        if '@raw' in self.obj_args or '@owl' in self.obj_args:  # TODO(dromer): deprecate @owl on next stable release
            try:
                pd_raw_args = parse_pd_raw_args(self.obj_args)
                self.__attributes.update(pd_raw_args)
                self.__extern_type = "param"  # make sure output code is generated
            except PdRawException as e:
                self.add_error(str(e))

    def validate_configuration(self) -> None:
        if len(self.obj_args) == 0:
            self.add_warning(
                f"No name was given to this {self.obj_type} object. "
                "It should have a name to reduce the risk of errors.",
                NotificationEnum.WARNING_USELESS_OBJECT)
        if len(self._inlet_connections.get("0", [])) == 0:
            self.add_warning(
                "This object has no inlet connections. "
                "It does nothing and will be removed.",
                NotificationEnum.WARNING_USELESS_OBJECT)
        if self.obj_type in {"s", "send"} and len(self._inlet_connections.get("1", [])) > 0:
            self.add_error(
                "Connections to the right inlet of a send object "
                "are not supported. A name should be given.",
                NotificationEnum.ERROR_MISSING_REQUIRED_ARGUMENT)

    def to_hv(self) -> Dict:
        # note: control rate send/receive objects should not modify their name argument
        names = {
            "s": "",
            "send": "",
            "s~": "sndrcv_sig_",
            "send~": "sndrcv_sig_",
            "throw~": "thrwctch_sig_"
        }

        return {
            "type": "send",
            "args": {
                "name": names[self.obj_type] + self.__send_name,
                "extern": self.__extern_type,
                "attributes": self.__attributes,
            },
            "properties": {
                "x": self.pos_x,
                "y": self.pos_y
            },
            "annotations": {
                "scope": "public"
            }
        }



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023-2024 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
from typing import Dict, Optional, TYPE_CHECKING

from .HeavyIrObject import HeavyIrObject

if TYPE_CHECKING:
    from .HeavyGraph import HeavyGraph


class HIrReceive(HeavyIrObject):
    """ A specific implementation of the __receive object.
    """

    def __init__(
        self,
        obj_type: str,
        args: Optional[Dict] = None,
        graph: Optional['HeavyGraph'] = None,
        annotations: Optional[Dict] = None
    ) -> None:
        super().__init__("__receive", args=args, graph=graph, annotations=annotations)
        if args is not None and args["extern"]:
            # externed receivers must contain only alphanumeric characters or underscores,
            # so that the names can be easily and transparently turned into code
            if re.search(r"\W", args["name"]):
                self.add_error(f"Parameter and Event names may only contain \
                                alphanumeric characters or underscore: '{args['name']}'")



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2021-2024 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import jinja2
import os
import shutil
import time
from typing import Optional

from ..copyright import copyright_manager
from ..filters import filter_string_cap, filter_templates, filter_xcode_build, filter_xcode_fileref

from hvcc.interpreters.pd2hv.NotificationEnum import NotificationEnum
from hvcc.types.compiler import Generator, CompilerResp, CompilerNotif, CompilerMsg, ExternInfo
from hvcc.types.meta import Meta


class c2unity(Generator):
    """Generates a Audio Native Plugin wrapper for Unity 5.
    """

    @classmethod
    def compile(
        cls,
        c_src_dir: str,
        out_dir: str,
        externs: ExternInfo,
        patch_name: Optional[str] = None,
        patch_meta: Meta = Meta(),
        num_input_channels: int = 0,
        num_output_channels: int = 0,
        copyright: Optional[str] = None,
        verbose: Optional[bool] = False
    ) -> CompilerResp:

        tick = time.time()

        parameter_list = externs.parameters.inParam
        event_list = externs.events.inEvent
        table_list = externs.tables

        out_dir = os.path.join(out_dir, "unity")
        patch_name = patch_name or "heavy"

        copyright = copyright_manager.get_copyright_for_c(copyright)

        # initialise the jinja template environment
        env = jinja2.Environment()
        env.filters["xcode_build"] = filter_xcode_build
        env.filters["xcode_fileref"] = filter_xcode_fileref
        env.filters["cap"] = filter_string_cap
        env.loader = jinja2.FileSystemLoader(
            encoding="utf-8-sig",
            searchpath=[os.path.join(os.path.dirname(os.path.abspath(__file__)), "templates")])

        static_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "static")
        src_out_dir = os.path.join(out_dir, "source")

        try:
            # ensure that the output directory does not exist
            out_dir = os.path.abspath(out_dir)
            if os.path.exists(out_dir):
                shutil.rmtree(out_dir)

            # copy over static files
            shutil.copytree(static_dir, out_dir)

            # copy over generated C source files
            src_out_dir = os.path.join(out_dir, "source", "heavy")
            shutil.copytree(c_src_dir, src_out_dir)

            # generate files from templates
            for f in env.list_templates(filter_func=filter_templates):
                file_path = os.path.join(out_dir, f)
                file_path = file_path.replace("{{name}}", patch_name)

                if not os.path.exists(os.path.dirname(file_path)):
                    os.makedirs(os.path.dirname(file_path))

                with open(file_path, "w") as g:
                    g.write(env.get_template(f).render(
                        patch_name=patch_name,
                        files=os.listdir(src_out_dir),
                        num_input_channels=num_input_channels,
                        num_output_channels=num_output_channels,
                        parameters=parameter_list,
                        events=event_list,
                        tables=table_list,
                        pool_sizes_kb=externs.memoryPoolSizesKb,
                        compile_files=os.listdir(src_out_dir),
                        copyright=copyright))

            return CompilerResp(
                stage="c2unity",
                in_dir=c_src_dir,
                out_dir=out_dir,
                compile_time=time.time() - tick
            )

        except Exception as e:
            return CompilerResp(
                stage="c2unity",
                notifs=CompilerNotif(
                    has_error=True,
                    exception=e,
                    warnings=[],
                    errors=[CompilerMsg(
                        enum=NotificationEnum.ERROR_EXCEPTION,
                        message=str(e)
                    )]
                ),
                in_dir=c_src_dir,
                out_dir=out_dir,
                compile_time=time.time() - tick
            )



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023-2024 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Dict, List

from .HeavyObject import HeavyObject

from hvcc.types.IR import IRSignalList


class SignalVar(HeavyObject):

    __OPERATION_DICT = {
        "__var~f": "sVarf",
        "__var~i": "sVari",
        "__var_k~f": "__hv_var_k_f",
        "__var_k~i": "__hv_var_k_i",
        "__varwrite~f": "__hv_varwrite_f",
        "__varwrite~i": "__hv_varwrite_i",
        "__varread~f": "__hv_varread_f",
        "__varread~i": "__hv_varread_i"
    }

    @classmethod
    def get_C_struct(cls, obj_type: str = "") -> str:
        if obj_type == "__var~f":
            return "SignalVarf"
        elif obj_type == "__var~i":
            return "SignalVari"
        else:
            raise Exception()

    @classmethod
    def get_preamble(cls, obj_type: str) -> str:
        return cls.__OPERATION_DICT[obj_type]

    @classmethod
    def handles_type(cls, obj_type: str) -> bool:
        """Returns true if the object type can be handled by this class
        """
        return obj_type in cls.__OPERATION_DICT

    @classmethod
    def get_C_header_set(cls) -> set:
        return {"HvSignalVar.h"}

    @classmethod
    def get_C_file_set(cls) -> set:
        return {"HvSignalVar.h", "HvSignalVar.c"}

    @classmethod
    def get_C_init(cls, obj_type: str, obj_id: str, args: Dict) -> List[str]:
        assert obj_type in ["__var~f", "__var~i"], obj_type
        return [
            "{0}_init(&{0}_{1}, {2}, {3}, {4});".format(
                SignalVar.__OPERATION_DICT[obj_type],
                obj_id,
                f"{float(args['k'])}f" if obj_type.endswith("f") else int(args["k"]),
                f"{float(args['step'])}f" if obj_type.endswith("f") else int(args["step"]),
                "true" if args["reverse"] else "false")]

    @classmethod
    def get_C_free(cls, obj_type: str, obj_id: str, args: Dict) -> List[str]:
        return []

    @classmethod
    def get_C_onMessage(cls, obj_type: str, obj_id: str, inlet_index: int, args: Dict) -> List[str]:
        assert obj_type in ["__var~f", "__var~i"]
        return [
            "{0}_onMessage(_c, &Context(_c)->{0}_{1}, m);".format(
                cls.__OPERATION_DICT[obj_type],
                obj_id)]

    @classmethod
    def get_C_process(cls, process_dict: IRSignalList, obj_type: str, obj_id: str, args: Dict) -> List[str]:
        fmt = obj_type[-1]
        if obj_type in ["__var~f", "__var~i"]:
            # NOTE(mhroth): signal rate variables do not process anything
            return []
        elif obj_type in ["__varwrite~f", "__varwrite~i"]:
            return [
                "__hv_varwrite_{1}(&sVar{1}_{0}, VI{1}({2}));".format(
                    args["var_id"],
                    fmt,
                    cls._c_buffer(process_dict.inputBuffers[0])
                )]
        elif obj_type in ["__var_k~f", "__var_k~i"]:
            if args["k"] == 0.0 and args.get("step", 0.0) == 0.0:
                return ["__hv_zero_{0}(VO{0}({1}));".format(
                    fmt,
                    cls._c_buffer(process_dict.outputBuffers[0]))]
            else:
                c = [float(args["k"] + i * args.get("step", 0.0)) for i in range(8)]
                cx = ", ".join(["{0}f".format(f) for f in c]) if fmt == "f" else ", ".join([str(int(i)) for i in c])
                return ["__hv_var_k_{0}{3}(VO{0}({1}), {2});".format(
                    fmt,
                    cls._c_buffer(process_dict.outputBuffers[0]),
                    cx,
                    "_r" if args.get("reverse", False) else "")]
        elif obj_type in ["__varread~f", "__varread~i"]:
            return [
                "__hv_varread_{1}(&sVar{1}_{0}, VO{1}({2}));".format(
                    args["var_id"],
                    fmt,
                    cls._c_buffer(process_dict.outputBuffers[0])
                )]
        else:
            raise Exception("")



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023-2024 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Dict, List

from .HeavyObject import HeavyObject


class HeavyTable(HeavyObject):
    """Outputs code for the table object.
    """

    c_struct = "HvTable"
    preamble = "hTable"

    @classmethod
    def get_C_header_set(self) -> set:
        return {"HvTable.h"}

    @classmethod
    def get_C_file_set(self) -> set:
        return {"HvTable.h", "HvTable.c"}

    @classmethod
    def get_table_data_decl(cls, obj_type: str, obj_id: str, args: Dict) -> List[str]:
        if len(args.get("values", [])) > 0:
            return [
                "float hTable_{0}_data[{1}] = {{{2}}};".format(
                    obj_id,
                    len(args["values"]),
                    ", ".join([f"{float(v)}f" for v in args["values"]]))]
        else:
            return []

    @classmethod
    def get_C_init(cls, obj_type: str, obj_id: str, args: Dict) -> List[str]:
        if len(args.get("values", [])) > 0:
            return [
                "hTable_initWithData(&hTable_{0}, {1}, hTable_{0}_data);".format(
                    obj_id,
                    len(args["values"]))]
        else:
            return [
                "hTable_init(&hTable_{0}, {1});".format(
                    obj_id,
                    int(args.get("size", 256)))]  # 1KB default memory allocation

    @classmethod
    def get_C_free(cls, obj_type: str, obj_id: str, args: Dict) -> List[str]:
        return ["{0}_free(&{0}_{1});".format(
            cls.preamble,
            obj_id)]

    @classmethod
    def get_C_onMessage(cls, obj_type: str, obj_id: str, inlet_index: int, args: Dict) -> List[str]:
        return [
            f"hTable_onMessage(_c, &Context(_c)->hTable_{obj_id}, {inlet_index}, m, &hTable_{obj_id}_sendMessage);"
        ]



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023-2024 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Dict, List

from .HeavyObject import HeavyObject

from hvcc.types.IR import IRSignalList


class SignalLine(HeavyObject):

    c_struct = "SignalLine"
    preamble = "sLine"

    @classmethod
    def get_C_header_set(cls) -> set:
        return {"HvSignalLine.h"}

    @classmethod
    def get_C_file_set(cls) -> set:
        return {"HvSignalLine.h", "HvSignalLine.c"}

    @classmethod
    def get_C_init(cls, obj_type: str, obj_id: str, args: Dict) -> List[str]:
        return [
            f"sLine_init(&sLine_{obj_id});"
        ]

    @classmethod
    def get_C_free(cls, obj_type: str, obj_id: str, args: Dict) -> List[str]:
        return []  # nothing to free

    @classmethod
    def get_C_onMessage(cls, obj_type: str, obj_id: str, inlet_index: int, args: Dict) -> List[str]:
        return [
            f"sLine_onMessage(_c, &Context(_c)->sLine_{obj_id}, {inlet_index}, m, NULL);"
        ]

    @classmethod
    def get_C_process(cls, process_dict: IRSignalList, obj_type: str, obj_id: str, args: Dict) -> List[str]:
        return [
            "__hv_line_f(&sLine_{0}, VOf({1}));".format(
                process_dict.id,
                cls._c_buffer(process_dict.outputBuffers[0])
            )
        ]



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from typing import List


class PrettyfyC:

    @classmethod
    def prettyfy_file(
        cls,
        file_in: str,
        file_out: str,
        indent: int = 0,
        step: int = 2,
        delete_input_on_finish: bool = False
    ) -> None:
        with open(file_in, "r") as f:
            with open(file_out, "w") as g:
                for line in f:
                    indent -= line.count("}")
                    new_line = (" " * (step * indent)) + line
                    g.write(new_line + os.linesep)
                    indent += line.count("{")

        if delete_input_on_finish:
            os.remove(file_in)

    @classmethod
    def prettyfy_list(
        cls,
        list_in: List,
        indent: int = 0,
        step: int = 2
    ) -> List:
        g = []
        for line in list_in:
            indent -= line.count("}")
            new_line = (" " * (step * indent)) + line
            g.append(new_line)
            indent += line.count("{")
        return g



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional, Dict

from .HeavyLangObject import HeavyLangObject
from .HeavyIrObject import HeavyIrObject
from .HeavyGraph import HeavyGraph


class HLangUnop(HeavyLangObject):

    __HEAVY_DICT = {
        "log": ["__log"],
        "log10": ["__log10"],
        "log2": ["__log2", "__log2~f"],
        "cos": ["__cos", "__cos~f"],
        "acos": ["__acos", "__acos~f"],
        "cosh": ["__cosh", "__cosh~f"],
        "acosh": ["__acosh", "__acosh~f"],
        "sin": ["__sin", "__sin~f"],
        "asin": ["__asin", "__asin~f"],
        "sinh": ["__sinh", "__sinh~f"],
        "asinh": ["__asinh", "__asin~f"],
        "tan": ["__tan", "__tan~f"],
        "atan": ["__atan", "__atan~f"],
        "tanh": ["__tanh", "__tanh~f"],
        "atanh": ["__atanh", "__atanh~f"],
        "exp": ["__exp", "__exp~f"],
        "sqrt": ["__sqrt", "__sqrt~f"],
        "abs": ["__abs", "__abs~f", "__abs~i"],
        "floor": ["__floor", "__floor~f"],
        "ceil": ["__ceil", "__ceil~f"],
        "cast_fi": ["__cast~fi"],
        "cast_if": ["__cast~if"]
    }

    def __init__(
        self,
        obj_type: str,
        args: Dict,
        graph: 'HeavyGraph',
        annotations: Optional[Dict] = None
    ) -> None:
        assert self.handles_type(obj_type)
        super().__init__(obj_type, args, graph,
                         num_inlets=1,
                         num_outlets=1,
                         annotations=annotations)

    @classmethod
    def handles_type(cls, obj_type: str) -> bool:
        """ Returns True if this class handles the given object type. False otherwise.
        """
        return obj_type in HLangUnop.__HEAVY_DICT

    def reduce(self) -> Optional[tuple]:
        if len(self.inlet_connections[0]) == 0:
            self.add_error("Unary operator objects must have an input.")

        # TODO(mhroth): there is no general consensus on how to handle incoming
        # and outgoing connections of different types, given the local let-type
        # definitions.

        if self.type == "cast_fi":
            x = HeavyIrObject(self.type)  # is this correct? no idea what x is otherwise
            return ({HeavyIrObject("__cast~fi", {})}, self.get_connection_move_list(x))
        elif self.type == "cast_if":
            x = HeavyIrObject(self.type)  # is this correct? no idea what x is otherwise
            return ({HeavyIrObject("__cast~if", {})}, self.get_connection_move_list(x))
        elif self.has_inlet_connection_format("f"):
            ir_type = f"__{self.type}~f"
            assert ir_type in HLangUnop.__HEAVY_DICT[self.type]
            x = HeavyIrObject(ir_type)
            return ({x}, self.get_connection_move_list(x, "~f>"))
        elif self.has_inlet_connection_format("i"):
            ir_type = f"__{self.type}~i"
            assert ir_type in HLangUnop.__HEAVY_DICT[self.type]
            x = HeavyIrObject(ir_type)
            return ({x}, self.get_connection_move_list(x, "~i>"))
        elif self.has_inlet_connection_format("c"):
            ir_type = f"__{self.type}"
            assert ir_type in HLangUnop.__HEAVY_DICT[self.type]
            x = HeavyIrObject(ir_type)
            return ({x}, self.get_connection_move_list(x, "-->"))
        else:
            self.add_error("Unknown inlet configuration.")
            return None



# Copyright (C) 2021-2024 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import shutil
import time
import jinja2
from typing import Optional

from ..copyright import copyright_manager
from ..filters import filter_uniqueid

from hvcc.interpreters.pd2hv.NotificationEnum import NotificationEnum
from hvcc.types.compiler import Generator, CompilerResp, CompilerMsg, CompilerNotif, ExternInfo
from hvcc.types.meta import Meta, DPF


class c2dpf(Generator):
    """ Generates a DPF wrapper for a given patch.
    """

    @classmethod
    def compile(
        cls,
        c_src_dir: str,
        out_dir: str,
        externs: ExternInfo,
        patch_name: Optional[str] = None,
        patch_meta: Meta = Meta(),
        num_input_channels: int = 0,
        num_output_channels: int = 0,
        copyright: Optional[str] = None,
        verbose: Optional[bool] = False
    ) -> CompilerResp:

        tick = time.time()

        out_dir = os.path.join(out_dir, "plugin")
        receiver_list = externs.parameters.inParam
        sender_list = externs.parameters.outParam

        dpf_meta: DPF = patch_meta.dpf
        dpf_path = dpf_meta.dpf_path

        copyright_c = copyright_manager.get_copyright_for_c(copyright)

        try:
            # ensure that the output directory does not exist
            out_dir = os.path.abspath(out_dir)
            if os.path.exists(out_dir):
                shutil.rmtree(out_dir)

            # copy over static files
            shutil.copytree(os.path.join(os.path.dirname(__file__), "static"), out_dir)
            shutil.copy(os.path.join(os.path.dirname(__file__), "static/README.md"), f'{out_dir}/../')

            # copy over generated C source files
            source_dir = os.path.join(out_dir, "source")
            shutil.copytree(c_src_dir, source_dir)

            # initialize the jinja template environment
            env = jinja2.Environment()
            env.filters["uniqueid"] = filter_uniqueid

            env.loader = jinja2.FileSystemLoader(
                os.path.join(os.path.dirname(os.path.abspath(__file__)), "templates"))

            # generate DPF wrapper from template
            dpf_h_path = os.path.join(source_dir, f"HeavyDPF_{patch_name}.hpp")
            with open(dpf_h_path, "w") as f:
                f.write(env.get_template("HeavyDPF.hpp").render(
                    name=patch_name,
                    meta=dpf_meta,
                    class_name=f"HeavyDPF_{patch_name}",
                    num_input_channels=num_input_channels,
                    num_output_channels=num_output_channels,
                    receivers=receiver_list,
                    senders=sender_list,
                    copyright=copyright_c))
            dpf_cpp_path = os.path.join(source_dir, f"HeavyDPF_{patch_name}.cpp")
            with open(dpf_cpp_path, "w") as f:
                f.write(env.get_template("HeavyDPF.cpp").render(
                    name=patch_name,
                    meta=dpf_meta,
                    class_name=f"HeavyDPF_{patch_name}",
                    num_input_channels=num_input_channels,
                    num_output_channels=num_output_channels,
                    receivers=receiver_list,
                    senders=sender_list,
                    pool_sizes_kb=externs.memoryPoolSizesKb,
                    copyright=copyright_c))
            if dpf_meta.enable_ui:
                dpf_ui_path = os.path.join(source_dir, f"HeavyDPF_{patch_name}_UI.cpp")
                with open(dpf_ui_path, "w") as f:
                    f.write(env.get_template("HeavyDPF_UI.cpp").render(
                        name=patch_name,
                        meta=dpf_meta,
                        class_name=f"HeavyDPF_{patch_name}",
                        receivers=receiver_list,
                        senders=sender_list,
                        copyright=copyright_c))
            dpf_h_path = os.path.join(source_dir, "DistrhoPluginInfo.h")
            with open(dpf_h_path, "w") as f:
                f.write(env.get_template("DistrhoPluginInfo.h").render(
                    name=patch_name,
                    meta=dpf_meta,
                    class_name=f"HeavyDPF_{patch_name}",
                    num_input_channels=num_input_channels,
                    num_output_channels=num_output_channels,
                    pool_sizes_kb=externs.memoryPoolSizesKb,
                    copyright=copyright_c))

            # plugin makefile
            with open(os.path.join(source_dir, "Makefile"), "w") as f:
                f.write(env.get_template("Makefile_plugin").render(
                    name=patch_name,
                    meta=dpf_meta,
                    nosimd=patch_meta.nosimd,
                    dpf_path=dpf_path))

            # project makefile
            with open(os.path.join(source_dir, "../../Makefile"), "w") as f:
                f.write(env.get_template("Makefile_project").render(
                    name=patch_name,
                    meta=dpf_meta,
                    dpf_path=dpf_path))

            return CompilerResp(
                stage="c2dpf",
                in_dir=c_src_dir,
                out_dir=out_dir,
                out_file=os.path.basename(dpf_h_path),
                compile_time=time.time() - tick
            )

        except Exception as e:
            return CompilerResp(
                stage="c2dpf",
                notifs=CompilerNotif(
                    has_error=True,
                    exception=e,
                    warnings=[],
                    errors=[CompilerMsg(
                        enum=NotificationEnum.ERROR_EXCEPTION,
                        message=str(e)
                    )]
                ),
                in_dir=c_src_dir,
                out_dir=out_dir,
                compile_time=time.time() - tick
            )



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from .HeavyIrObject import HeavyIrObject


class Connection:
    """ A Connection describes a connection between two objects.
    """

    def __init__(
        self,
        from_object: 'HeavyIrObject',
        outlet_index: int,
        to_object: 'HeavyIrObject',
        inlet_index: int,
        conn_type: str
    ) -> None:
        self.from_object = from_object
        self.outlet_index = outlet_index
        self.to_object = to_object
        self.inlet_index = inlet_index
        self.type = conn_type

        # cache the hash of this object
        self.__hash = hash((
            self.from_object, self.outlet_index,
            self.to_object, self.inlet_index,
            self.type))

    def copy(
        self,
        from_object: Optional['HeavyIrObject'] = None,
        outlet_index: Optional[int] = None,
        to_object: Optional['HeavyIrObject'] = None,
        inlet_index: Optional[int] = None,
        type: Optional[str] = None
    ) -> 'Connection':
        """ Create a new connection based on the existing one, changing the given values.
        """
        return Connection(from_object=self.from_object if from_object is None else from_object,
                          outlet_index=self.outlet_index if outlet_index is None else outlet_index,
                          to_object=self.to_object if to_object is None else to_object,
                          inlet_index=self.inlet_index if inlet_index is None else inlet_index,
                          conn_type=self.type if type is None else type)

    @property
    def is_signal(self) -> bool:
        return self.is_signal_type(self.type)

    @property
    def is_control(self) -> bool:
        return self.type == "-->"

    @property
    def is_float_signal(self) -> bool:
        return self.type == "~f>"

    @property
    def is_integer_signal(self) -> bool:
        return self.type == "~i>"

    @property
    def is_mixed(self) -> bool:
        return self.type == "-~>"

    @classmethod
    def is_signal_type(cls, type: Optional[str]) -> bool:
        return type in {"~i>", "~f>"}

    def __eq__(self, other: object) -> bool:
        return self.__hash == other.__hash__() if isinstance(other, Connection) else False

    def __hash__(self) -> int:
        return self.__hash

    def __repr__(self) -> str:
        return f"[{self.from_object}:{self.outlet_index}] {self.type} [{self.to_object}:{self.inlet_index}]"



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023-2024 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
from typing import Dict, List, Optional

from .HeavyIrObject import HeavyIrObject
from .HeavyGraph import HeavyGraph

from hvcc.types.IR import IRSendMessage


class HIrSend(HeavyIrObject):
    """ A specific implementation of the __send object.
    """

    def __init__(
        self,
        obj_type: str,
        args: Optional[Dict] = None,
        graph: Optional[HeavyGraph] = None,
        annotations: Optional[Dict] = None
    ) -> None:
        super().__init__("__send", args=args, graph=graph, annotations=annotations)
        if args is not None and args["extern"]:
            # output parameters must contain only alphanumeric characters or underscores,
            # so that the names can be easily and transparently turned into code
            if re.search(r"\W", args["name"]):
                self.add_error(f"Parameter and Event names may only contain \
                                alphanumeric characters or underscore: '{args['name']}'")

    def get_ir_control_list(self) -> List[IRSendMessage]:
        if self.graph is not None and self.name is not None:
            receive_objs = self.graph.resolve_objects_for_name(self.name, "__receive")
            on_message_list = [x for o in receive_objs for x in o.get_ir_on_message(inlet_index=0)]

            return [IRSendMessage(
                id=self.id,
                type="send",
                onMessage=[on_message_list],
                extern=self.args["extern"],
                attributes=self.args["attributes"],
                hash=self.args["hash"],
                display=self.args["name"],
                name=((f"_{self.args['name']}") if re.match(r"\d", self.args["name"]) else self.args["name"])
            )]
        else:
            return []



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023-2024 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Callable, Dict, List

from .HeavyObject import HeavyObject

from hvcc.types.IR import IROnMessage, IRObjectdict


class ControlDelay(HeavyObject):

    c_struct = "ControlDelay"
    preamble = "cDelay"

    @classmethod
    def get_C_header_set(cls) -> set:
        return {"HvControlDelay.h"}

    @classmethod
    def get_C_file_set(cls) -> set:
        return {"HvControlDelay.h", "HvControlDelay.c"}

    @classmethod
    def get_C_init(cls, obj_type: str, obj_id: str, args: Dict) -> List[str]:
        return [
            "{0}_init(this, &{0}_{1}, {2}f);".format(
                cls.preamble,
                obj_id,
                float(args["delay"]))
        ]

    @classmethod
    def get_C_free(cls, obj_type: str, obj_id: str, args: Dict) -> List[str]:
        return []  # no need to free any control binop objects

    @classmethod
    def get_C_onMessage(cls, obj_type: str, obj_id: str, inlet_index: int, args: Dict) -> List[str]:
        return [
            "{0}_onMessage(_c, &Context(_c)->{0}_{1}, {2}, m, "
            "&{0}_{1}_sendMessage);".format(
                cls.preamble,
                obj_id,
                inlet_index)
        ]

    @classmethod
    def get_C_impl(
        cls,
        obj_type: str,
        obj_id: str,
        on_message_list: List[List[IROnMessage]],
        get_obj_class: Callable,
        objects: Dict[str, IRObjectdict]
    ) -> List[str]:
        send_message_list = [
            f"cDelay_{obj_id}_sendMessage(HeavyContextInterface *_c, int letIn, const HvMessage *const m) {{"
        ]
        send_message_list.append(f"cDelay_clearExecutingMessage(&Context(_c)->cDelay_{obj_id}, m);")
        send_message_list.extend(
            cls._get_on_message_list(on_message_list[0], get_obj_class, objects))
        send_message_list.append("}")  # end function
        return send_message_list



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import List, Dict


class PdRawException(Exception):
    pass


def replace_owl(args: List) -> List:
    new_args = []
    for arg in args:
        new_arg = arg.replace('owl', 'raw')
        new_args.append(new_arg)
    return new_args


def parse_pd_raw_args(args: List) -> Dict:
    """Parses a list of puredata send or receive objects looking for @raw and legacy @owl*
    annotations, parsing everything and throwing errors when syntax is not
    correct or values are of incorrect type"""

    attrdict = {}

    # define default values
    attrdict["min"] = 0.0
    attrdict["max"] = 1.0

    args = replace_owl(args)  # TODO(dromer): deprecate @owl on next stable release

    for raw_param in {'@raw', '@raw_min', '@raw_max', '@raw_default', '@raw_param'}:
        if raw_param not in args:
            continue

        i = args.index(raw_param)

        if raw_param in {'@raw', '@raw_param'}:
            try:
                attrdict["raw"] = args[i + 1]
            except IndexError:
                raise PdRawException(f"{raw_param} annotation missing assigned parameter")
            if raw_param == '@raw':
                try:
                    # expect the presence of up to 3 parameters which can be converted to float
                    attrdict["min"] = float(args[i + 2])
                    attrdict["max"] = float(args[i + 3])
                    attrdict["default"] = float(args[i + 4])
                except (IndexError, ValueError):
                    # otherwise keep default
                    pass
        elif raw_param in {'@raw_min', '@raw_max', '@raw_default'}:
            # make sure that it is a float value
            try:
                attrdict[raw_param.split('@raw_')[1]] = float(args[i + 1])
            except ValueError:
                raise PdRawException(f"{raw_param} annotation value '{args[i + 1]}' is not numeric")
            except IndexError:
                raise PdRawException(f"{raw_param} annotation is missing its value")

    if attrdict.get("default") is None:
        attrdict["default"] = (attrdict["max"] - attrdict["min"]) / 2.0

    return attrdict



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import os
from typing import Optional


def get_default_copyright_text() -> str:
    with open(os.path.join(os.path.dirname(__file__), "default_template.txt"), "r") as f:
        copyright = f.read().format(datetime.datetime.now().year)
    return copyright


def get_copyright_for_c(copyright_text: Optional[str] = None) -> str:
    """ Returns an input string as a C-formatted comment, otherwise a default
        copyright statement if the input is None.
    """
    return comment_for_c(copyright_text or get_default_copyright_text())


def get_copyright_for_xml(copyright_text: Optional[str] = None) -> str:
    """ Returns an input string as an xml comment, otherwise a default
        copyright statement if the input is None.
    """
    return comment_for_xml(copyright_text or get_default_copyright_text())


def comment_for_c(comment: str) -> str:
    """ Returns the input string as a C-formatted comment, suitable for copyright statements.
    """
    if "\n" in comment:
        return "/**\n * {0}\n */".format("\n * ".join(comment.split("\n")))
    else:
        return f"/** {comment} */"


def comment_for_xml(comment: str) -> str:
    """ Returns the input string as a XML comment, suitable for copyright statements.
    """
    if "\n" in comment:
        return f"<!--\n{comment}\n-->"
    else:
        return f"<!-- {comment} -->"



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023-2024 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import json
import os
import time
from typing import List, Optional

from hvcc.interpreters.pd2hv.PdParser import PdParser
from hvcc.types.compiler import CompilerResp, CompilerNotif


class Colours:
    purple = "\033[95m"
    cyan = "\033[96m"
    dark_cyan = "\033[36m"
    blue = "\033[94m"
    green = "\033[92m"
    yellow = "\033[93m"
    red = "\033[91m"
    bold = "\033[1m"
    underline = "\033[4m"
    end = "\033[0m"


class pd2hv:

    @classmethod
    def get_supported_objects(cls) -> List:
        return PdParser.get_supported_objects()

    @classmethod
    def compile(
        cls,
        pd_path: str,
        hv_dir: str,
        search_paths: Optional[List] = None,
        verbose: bool = False,
        export_args: bool = False
    ) -> CompilerResp:

        tick = time.time()

        parser = PdParser()  # create parser state
        if search_paths is not None:
            for p in search_paths:
                parser.add_absolute_search_directory(p)

        pd_graph = parser.graph_from_file(pd_path)
        notices = pd_graph.get_notices()

        # check for errors
        if len(notices.errors) > 0:
            return CompilerResp(
                stage="pd2hv",
                obj_counter=parser.obj_counter,
                notifs=CompilerNotif(
                    has_error=True,
                    errors=notices.errors,
                    warnings=notices.warnings
                ),
                in_dir=os.path.dirname(pd_path),
                in_file=os.path.basename(pd_path),
                compile_time=(time.time() - tick)
            )

        if not os.path.exists(hv_dir):
            os.makedirs(hv_dir)

        hv_file = f"{os.path.splitext(os.path.basename(pd_path))[0]}.hv.json"
        hv_path = os.path.join(hv_dir, hv_file)
        with open(hv_path, "w") as f:
            if verbose:
                json.dump(pd_graph.to_hv(export_args=export_args),
                          f,
                          sort_keys=True,
                          indent=2,
                          separators=(",", ": "))
            else:
                json.dump(pd_graph.to_hv(export_args=export_args), f)

        return CompilerResp(
            stage="pd2hv",
            obj_counter=parser.obj_counter,
            notifs=CompilerNotif(
                warnings=notices.warnings
            ),
            in_dir=os.path.dirname(pd_path),
            in_file=os.path.basename(pd_path),
            out_dir=hv_dir,
            out_file=hv_file,
            compile_time=(time.time() - tick)
        )


def main() -> None:
    parser = argparse.ArgumentParser(
        description="Converts a Pd patch into the Heavy language format.")
    parser.add_argument(
        "pd_path",
        help="The Pd patch to convert to Heavy.")
    parser.add_argument(
        "hv_dir",
        help="Directory to store generated Heavy patches.")
    parser.add_argument(
        "--export",
        help="Export Heavy arguments. Use this to make precompiled patches.",
        action="count")
    parser.add_argument(
        "-v",
        "--verbose",
        help="Show debugging information.",
        action="count")
    args = parser.parse_args()

    args.pd_path = os.path.abspath(os.path.expanduser(args.pd_path))
    args.hv_dir = os.path.abspath(os.path.expanduser(args.hv_dir))

    result = pd2hv.compile(
        pd_path=args.pd_path,
        hv_dir=args.hv_dir,
        search_paths=None,
        verbose=args.verbose,
        export_args=args.export)

    for i, n in enumerate(result.notifs.errors):
        print("{0:3d}) {1}Error #{2:4d}:{3} {4}".format(
            i + 1,
            Colours.red,
            n.enum,
            Colours.end,
            n.message))
    for i, n in enumerate(result.notifs.warnings):
        print("{0:3d}) {1}Warning #{2:4d}:{3} {4}".format(
            i + 1,
            Colours.yellow,
            n.enum,
            Colours.end,
            n.message))

    if args.verbose:
        if len(result.notifs.errors) == 0:
            print("Heavy file written to", os.path.join(result.out_dir, result.out_file))
        print("Total pd2hv compile time: {0:.2f}ms".format(result.compile_time * 1000))


if __name__ == "__main__":
    main()



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023-2024 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Dict, List

from .HeavyObject import HeavyObject

from hvcc.types.IR import IRSignalList


class SignalTabwrite(HeavyObject):
    """Handles __tabwrite~f
    """

    c_struct = "SignalTabwrite"
    preamble = "sTabwrite"

    @classmethod
    def get_C_header_set(cls) -> set:
        return {"HvSignalTabwrite.h"}

    @classmethod
    def get_C_file_set(cls) -> set:
        return {"HvSignalTabwrite.h", "HvSignalTabwrite.c"}

    @classmethod
    def get_C_free(cls, obj_type: str, obj_id: str, args: Dict) -> List[str]:
        return []

    @classmethod
    def get_C_init(cls, obj_type: str, obj_id: str, args: Dict) -> List[str]:
        return [f"sTabwrite_init(&sTabwrite_{obj_id}, &hTable_{args['table_id']});"]

    @classmethod
    def get_C_onMessage(cls, obj_type: str, obj_id: str, inlet_index: int, args: Dict) -> List[str]:
        return [f"sTabwrite_onMessage(_c, &Context(_c)->sTabwrite_{obj_id}, {inlet_index}, m, NULL);"]

    @classmethod
    def get_C_process(cls, process_dict: IRSignalList, obj_type: str, obj_id: str, args: Dict) -> List[str]:
        if obj_type == "__tabwrite~f":
            return [
                "__hv_tabwrite_f(&sTabwrite_{0}, {1});".format(
                    process_dict.id,
                    ", ".join([f"VIf({cls._c_buffer(b)})" for b in process_dict.inputBuffers])
                )]
        elif obj_type == "__tabwrite_stoppable~f":
            return [
                "__hv_tabwrite_stoppable_f(&sTabwrite_{0}, {1});".format(
                    process_dict.id,
                    ", ".join([f"VIf({cls._c_buffer(b)})" for b in process_dict.inputBuffers])
                )]
        else:
            raise Exception()



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023-2024 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from hvcc.types.compiler import CompilerNotif


class HeavyException(Exception):
    """ This exception is raised if anything goes wrong during the hv2ir compilation process.
    """

    def __init__(self, message: str = "") -> None:
        super(Exception, self).__init__(message)
        self.message = message
        self.notes: CompilerNotif = CompilerNotif()



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023-2024 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Dict, Optional

from .HeavyException import HeavyException
from .HeavyIrObject import HeavyIrObject
from .HeavyGraph import HeavyGraph

from hvcc.types.Lang import LangLetType


class HIrInlet(HeavyIrObject):
    """ A specific implementation of the inlet object.
    """

    def __init__(
        self,
        obj_type: str,
        args: Optional[Dict] = None,
        graph: Optional[HeavyGraph] = None,
        annotations: Optional[Dict] = None
    ) -> None:
        super().__init__("__inlet", args=args, graph=graph, annotations=annotations)

    def _resolved_outlet_type(self, outlet_index: int = 0) -> Optional[LangLetType]:
        if self.graph is not None:
            connections = self.graph.inlet_connections[self.args["index"]]
            connection_type_set = {c.type for c in connections}
            if len(connection_type_set) == 0:
                # object has no incident connections.
                return "-->"  # outlet type defaults to control (-->)
            elif len(connection_type_set) == 1:
                return list(connection_type_set)[0]
            else:
                raise HeavyException(
                    f"{self} has multiple incident connections of differing type. "
                    "The outlet type cannot be explicitly resolved.")
        return None



# Heavy Compiler Collection
# Copyright (C) 2024 Wasted Audio
#
# SPDX-License-Identifier: GPL-3.0-only

from abc import ABC, abstractmethod
from collections import Counter, defaultdict
from typing import Dict, List, Optional, Tuple

from pydantic import BaseModel, RootModel

from hvcc.interpreters.pd2hv.NotificationEnum import NotificationEnum
from hvcc.types.meta import Meta
from hvcc.types.IR import IRGraph, IRReceiver, IRSendMessage, IRTable


class CompilerMsg(BaseModel):
    enum: NotificationEnum = NotificationEnum.EMPTY
    message: str


class CompilerNotif(BaseModel, arbitrary_types_allowed=True):
    has_error: bool = False
    exception: Optional[Exception] = None
    warnings: List[CompilerMsg] = []
    errors: List[CompilerMsg] = []


class CompilerResp(BaseModel):
    stage: str
    notifs: CompilerNotif = CompilerNotif()
    in_dir: str = ""
    in_file: str = ""
    out_dir: str = ""
    out_file: str = ""
    compile_time: float = 0.0
    obj_counter: Counter = Counter()
    obj_perf: Dict[str, Counter] = defaultdict(Counter)
    ir: Optional[IRGraph] = None


class CompilerResults(RootModel):
    root: Dict[str, CompilerResp]


class ExternParams(BaseModel):
    inParam: List[Tuple[str, IRReceiver]] = []
    outParam: List[Tuple[str, IRSendMessage]] = []


class ExternEvents(BaseModel):
    inEvent: List[Tuple[str, IRReceiver]] = []
    outEvent: List[Tuple[str, IRSendMessage]] = []


class ExternMidi(BaseModel):
    inMidi: List[str] = []
    outMidi: List[str] = []


class ExternMemoryPool(BaseModel):
    internal: int = 0
    inputQueue: int = 0
    outputQueue: int = 0


class ExternInfo(BaseModel):
    parameters: ExternParams = ExternParams()
    events: ExternEvents = ExternEvents()
    midi: ExternMidi = ExternMidi()
    tables: List[Tuple[str, IRTable]] = []
    memoryPoolSizesKb: ExternMemoryPool = ExternMemoryPool()


class Generator(ABC):

    @classmethod
    @abstractmethod
    def compile(
        cls,
        c_src_dir: str,
        out_dir: str,
        externs: ExternInfo,
        patch_name: Optional[str] = None,
        patch_meta: Meta = Meta(),
        num_input_channels: int = 0,
        num_output_channels: int = 0,
        copyright: Optional[str] = None,
        verbose: Optional[bool] = False
    ) -> CompilerResp:
        raise NotImplementedError



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional, List, Dict

from .NotificationEnum import NotificationEnum
from .PdObject import PdObject


class PdAudioIoObject(PdObject):
    def __init__(
        self,
        obj_type: str,
        obj_args: Optional[List] = None,
        pos_x: int = 0,
        pos_y: int = 0
    ) -> None:
        assert obj_type in {"adc~", "dac~"}
        super().__init__(obj_type, obj_args, pos_x, pos_y)

    def validate_configuration(self) -> None:
        # ensure that only signal connections are made to the dac
        for i, connections in self._inlet_connections.items():
            if any(c.conn_type == "-->" for c in connections):
                self.add_error(
                    f"{self.obj_type} does not support control connections (inlet {i}). They should be removed.",
                    NotificationEnum.ERROR_UNSUPPORTED_CONNECTION)

    def to_hv(self) -> Dict:
        return {
            "type": self.obj_type.strip("~"),
            "args": {
                "channels": [1, 2] if len(self.obj_args) == 0 else [int(a) for a in self.obj_args]
            },
            "properties": {
                "x": self.pos_x,
                "y": self.pos_y
            }
        }



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023-2024 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Dict, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from .PdObject import PdObject


class Connection:
    def __init__(
        self,
        from_obj: 'PdObject',
        outlet_index: int,
        to_obj: 'PdObject',
        inlet_index: int,
        conn_type: Optional[str]  # not actually Optional. This is due to the requirement in HeavyObject.
    ) -> None:
        assert from_obj is not None
        assert to_obj is not None
        assert conn_type is not None

        self.__from_obj = from_obj
        self.__to_obj = to_obj
        self.__hv_json: Dict = {
            "from": {
                "id": from_obj.obj_id,
                "outlet": outlet_index
            },
            "to": {
                "id": to_obj.obj_id,
                "inlet": inlet_index
            },
            "type": conn_type
        }

    @property
    def from_obj(self) -> 'PdObject':
        return self.__from_obj

    @property
    def from_id(self) -> str:
        return self.__hv_json["from"]["id"]

    @property
    def outlet_index(self) -> int:
        return self.__hv_json["from"]["outlet"]

    @property
    def to_obj(self) -> 'PdObject':
        return self.__to_obj

    @property
    def to_id(self) -> str:
        return self.__hv_json["to"]["id"]

    @property
    def inlet_index(self) -> int:
        return self.__hv_json["to"]["inlet"]

    @property
    def conn_type(self) -> str:
        return self.__hv_json["type"]

    def to_hv(self) -> Dict:
        return self.__hv_json

    def __repr__(self) -> str:
        return "{0}:{1} {4} {2}:{3}".format(
            self.from_id,
            self.outlet_index,
            self.to_id,
            self.inlet_index,
            self.conn_type)



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional, List, Dict

from .PdObject import PdObject


class HvSwitchcase(PdObject):
    def __init__(
        self,
        obj_type: str,
        obj_args: Optional[List] = None,
        pos_x: int = 0,
        pos_y: int = 0
    ) -> None:
        assert obj_type == "__switchcase"
        super().__init__(obj_type, obj_args, pos_x, pos_y)

        # ensure that correct case hashes are generated
        for i, a in enumerate(self.obj_args):
            try:
                self.obj_args[i] = float(a)
            except Exception:
                pass

    def get_outlet_connection_type(self, outlet_index: int = 0) -> str:
        return "-->"

    def to_hv(self) -> Dict:
        return {
            "type": "__switchcase",
            "args": {
                "cases": self.obj_args
            },
            "properties": {
                "x": self.pos_x,
                "y": self.pos_y
            }
        }



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2021-2024 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import importlib
import inspect
import json
import os
import re
import sys
from typing import Any, List, Dict, Optional

from hvcc.interpreters.pd2hv import pd2hv
from hvcc.core.hv2ir import hv2ir
from hvcc.generators.ir2c import ir2c
from hvcc.generators.ir2c import ir2c_perf
from hvcc.generators.c2js import c2js
from hvcc.generators.c2daisy import c2daisy
from hvcc.generators.c2dpf import c2dpf
from hvcc.generators.c2owl import c2owl
from hvcc.generators.c2pdext import c2pdext
from hvcc.generators.c2wwise import c2wwise
from hvcc.generators.c2unity import c2unity
from hvcc.types.compiler import (
    CompilerResults, CompilerResp, CompilerNotif, CompilerMsg, Generator,
    ExternInfo, ExternMemoryPool, ExternMidi, ExternEvents, ExternParams
)
from hvcc.types.IR import IRGraph
from hvcc.types.meta import Meta


def add_error(
    results: CompilerResults,
    error: str
) -> CompilerResults:
    if "hvcc" in results.root:
        results.root["hvcc"].notifs.errors.append(CompilerMsg(message=error))
    else:
        results.root["hvcc"] = CompilerResp(
                                stage="hvcc",
                                notifs=CompilerNotif(
                                    has_error=True,
                                    errors=[CompilerMsg(message=error)]
                                )
                            )
    return results


def check_extern_name_conflicts(extern_type: str, extern_list: List, results: CompilerResults) -> None:
    """ In most of the generator code extern names become capitalised when used
        as enums. This method makes sure that there are no cases where two unique
        keys become the same after being capitalised.
        Note(joe): hvcc is probably the best place to check as at this point we
        have a list of all extern names.
    """
    for i, v in enumerate(extern_list):
        for j, u in enumerate(extern_list[i + 1:]):
            if v[0].upper() == u[0].upper():
                add_error(results,
                          f"Conflicting {extern_type} names '{v[0]}' and '{u[0]}', make sure that "
                          "capital letters are not the only difference.")


def count_midi_objects(hvir: IRGraph) -> Dict[str, List[str]]:
    in_midi = []
    out_midi = []

    midi_in_objs = [
        '__hv_bendin',
        '__hv_ctlin',
        '__hv_midiin',
        '__hv_midirealtimein',
        '__hv_notein',
        '__hv_pgmin',
        '__hv_polytouchin',
        '__hv_touchin',
    ]

    midi_out_objs = [
        '__hv_bendout',
        '__hv_ctlout',
        '__hv_midiout',
        '__hv_midioutport',
        '__hv_noteout',
        '__hv_pgmout',
        '__hv_polytouchout',
        '__hv_touchout',
    ]

    for recv in hvir.control.receivers.keys():
        if recv in midi_in_objs:
            in_midi.append(recv)

    for msg in hvir.control.sendMessage:
        if msg.name in midi_out_objs:
            out_midi.append(msg.name)

    return {
        'in': in_midi,
        'out': out_midi
    }


def filter_midi_from_out_parameters(output_parameter_list: List, midi_out_objects: List) -> List:
    new_out_list = []

    for item in output_parameter_list:
        if not item[0] in midi_out_objects:
            new_out_list.append(item)

    return new_out_list


def generate_extern_info(hvir: IRGraph, results: CompilerResults) -> ExternInfo:
    """ Simplifies the receiver/send and table lists by only containing values
        externed with @hv_param, @hv_event or @hv_table
    """
    # Exposed input parameters
    in_parameter_list = [(k, v) for k, v in hvir.control.receivers.items() if v.extern == "param"]
    in_parameter_list.sort(key=lambda x: x[0])
    check_extern_name_conflicts("input parameter", in_parameter_list, results)

    # Exposed input events
    in_event_list = [(k, v) for k, v in hvir.control.receivers.items() if v.extern == "event"]
    in_event_list.sort(key=lambda x: x[0])
    check_extern_name_conflicts("input event", in_event_list, results)

    # Exposed output parameters
    out_parameter_list = [(v.name, v) for v in hvir.control.sendMessage if v.extern == "param"]
    # remove duplicate output parameters/events
    # NOTE(joe): is the id argument important here? We'll only take the first one in this case.
    out_parameter_list = list(dict(out_parameter_list).items())
    out_parameter_list.sort(key=lambda x: x[0])
    check_extern_name_conflicts("output parameter", out_parameter_list, results)

    # Exposed output events
    out_event_list = [(v.name, v) for v in hvir.control.sendMessage if v.extern == "event"]
    out_event_list = list(dict(out_event_list).items())
    out_event_list.sort(key=lambda x: x[0])
    check_extern_name_conflicts("output event", out_event_list, results)

    # Exposed tables
    table_list = [(k, v) for k, v in hvir.tables.items() if v.extern]
    table_list.sort(key=lambda x: x[0])
    check_extern_name_conflicts("table", table_list, results)

    # Exposed midi objects
    midi_objects = count_midi_objects(hvir)

    # filter midi objects from the output parameters list
    out_parameter_list = filter_midi_from_out_parameters(out_parameter_list, midi_objects['out'])

    return ExternInfo(
        parameters=ExternParams(
            inParam=in_parameter_list,
            outParam=out_parameter_list
        ),
        events=ExternEvents(
            inEvent=in_event_list,
            outEvent=out_event_list
        ),
        midi=ExternMidi(
            inMidi=midi_objects['in'],
            outMidi=midi_objects['out']
        ),
        tables=table_list,
        # generate patch heuristics to ensure enough memory allocated for the patch
        memoryPoolSizesKb=ExternMemoryPool(
            internal=10,  # TODO(joe): should this increase if there are a lot of internal connections?
            inputQueue=max(2, int(
                len(in_parameter_list) +
                (len(in_event_list) / 4) +
                len(midi_objects['in'])  # TODO(dreamer): should this depend on the MIDI type?
            )),
            outputQueue=max(2, int(
                len(out_parameter_list) +
                (len(out_event_list) / 4) +
                len(midi_objects['out'])
            ))
        )
    )


def load_ext_generator(module_name: str, verbose: bool) -> Optional[Generator]:
    try:
        module = importlib.import_module(module_name)
        for _, member in inspect.getmembers(module):
            if inspect.isclass(member) and not inspect.isabstract(member) and issubclass(member, Generator):
                return member()
        if verbose:
            print(f"---> Module {module_name} does not contain a class derived from hvcc.types.Compiler")
        return None
    except ModuleNotFoundError:
        return None


def compile_dataflow(
    in_path: str,
    out_dir: str,
    patch_name: str = "heavy",
    patch_meta_file: Optional[str] = None,
    search_paths: Optional[List[str]] = None,
    generators: Optional[List[str]] = None,
    ext_generators: Optional[List[str]] = None,
    verbose: bool = False,
    copyright: Optional[str] = None,
    nodsp: Optional[bool] = False
) -> CompilerResults:
    results = CompilerResults(root={})
    patch_meta = Meta()

    # basic error checking on input
    if os.path.isfile(in_path):
        if not in_path.endswith((".pd")):
            return add_error(results, "Can only process Pd files.")
    elif os.path.isdir(in_path):
        if not os.path.basename("c"):
            return add_error(results, "Can only process c directories.")
    else:
        return add_error(results, f"Unknown input path {in_path}")

    # meta-data file
    if patch_meta_file:
        if os.path.isfile(patch_meta_file):
            with open(patch_meta_file) as json_file:
                try:
                    patch_meta_json = json.load(json_file)
                    patch_meta = Meta(**patch_meta_json)
                except Exception as e:
                    return add_error(results, f"Unable to open json_file: {e}")

    patch_name = patch_meta.name or patch_name
    generators = ["c"] if generators is None else [x.lower() for x in generators]

    if verbose:
        print("--> Generating C")
    results.root["pd2hv"] = pd2hv.pd2hv.compile(
        pd_path=in_path,
        hv_dir=os.path.join(out_dir, "hv"),
        search_paths=search_paths,
        verbose=verbose)

    # check for errors
    response: CompilerResp = list(results.root.values())[0]

    if response.notifs.has_error:
        return results

    subst_name = re.sub(r'\W', '_', patch_name)
    results.root["hv2ir"] = hv2ir.hv2ir.compile(
        hv_file=os.path.join(response.out_dir, response.out_file),
        # ensure that the ir filename has no funky characters in it
        ir_file=os.path.join(out_dir, "ir", f"{subst_name}.heavy.ir.json"),
        patch_name=patch_name,
        verbose=verbose)

    # check for errors
    if results.root["hv2ir"].notifs.has_error:
        return results

    # get the hvir data
    hvir = results.root["hv2ir"].ir
    assert hvir is not None
    patch_name = hvir.name.escaped
    externs = generate_extern_info(hvir, results)

    # get application path
    if getattr(sys, 'frozen', False) and hasattr(sys, '_MEIPASS'):
        application_path = os.path.join(sys._MEIPASS, 'hvcc')
    elif __file__:
        application_path = os.path.dirname(__file__)

    c_src_dir = os.path.join(out_dir, "c")
    results.root["ir2c"] = ir2c.ir2c.compile(
        hv_ir_path=os.path.join(results.root["hv2ir"].out_dir, results.root["hv2ir"].out_file),
        static_dir=os.path.join(application_path, "generators/ir2c/static"),
        output_dir=c_src_dir,
        externs=externs,
        copyright=copyright,
        nodsp=nodsp)

    # check for errors
    if results.root["ir2c"].notifs.has_error:
        return results

    # ir2c_perf
    results.root["ir2c_perf"] = CompilerResp(
        stage="ir2c_perf",
        obj_perf=ir2c_perf.ir2c_perf.perf(hvir, verbose=verbose),
        in_dir=results.root["hv2ir"].out_dir,
        in_file=results.root["hv2ir"].out_file,
    )

    # run the c2x generators, merge the results
    num_input_channels = hvir.signal.numInputBuffers
    num_output_channels = hvir.signal.numOutputBuffers

    gen_args: Dict[str, Any] = {
        'c_src_dir': c_src_dir,
        'out_dir': out_dir,
        'patch_name': patch_name,
        'patch_meta': patch_meta,
        'num_input_channels': num_input_channels,
        'num_output_channels': num_output_channels,
        'externs': externs,
        'copyright': copyright,
        'verbose': verbose
    }

    if "js" in generators:
        if verbose:
            print("--> Generating Javascript")
        results.root["c2js"] = c2js.c2js.compile(**gen_args)

    if "daisy" in generators:
        if verbose:
            print("--> Generating Daisy module")
        results.root["c2daisy"] = c2daisy.c2daisy.compile(**gen_args)

    if "dpf" in generators:
        if verbose:
            print("--> Generating DPF plugin")
        results.root["c2dpf"] = c2dpf.c2dpf.compile(**gen_args)

    if "owl" in generators:
        if verbose:
            print("--> Generating OWL plugin")
        results.root["c2owl"] = c2owl.c2owl.compile(**gen_args)

    if "pdext" in generators:
        if verbose:
            print("--> Generating Pd external")
        results.root["c2pdext"] = c2pdext.c2pdext.compile(**gen_args)

    if "unity" in generators:
        if verbose:
            print("--> Generating Unity plugin")
        results.root["c2unity"] = c2unity.c2unity.compile(**gen_args)

    if "wwise" in generators:
        if verbose:
            print("--> Generating Wwise plugin")
        results.root["c2wwise"] = c2wwise.c2wwise.compile(**gen_args)

    if ext_generators:
        for module_name in ext_generators:
            generator = load_ext_generator(module_name, verbose)
            if generator is not None:
                if verbose:
                    print(f"--> Executing custom generator from module {module_name}")
                results.root[module_name] = generator.compile(**gen_args)

    return results



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import argparse
import json

from hvcc.core.hv2ir.HeavyLangObject import HeavyLangObject
from hvcc.interpreters.pd2hv.PdParser import PdParser
from hvcc.types.meta import Meta, DPF, Daisy


gens = {
    "dpf": DPF,
    "daisy": Daisy
}


def main() -> None:
    parser = argparse.ArgumentParser(
        description="Some separate heavy utilities, wrapped into a single app")
    subparsers = parser.add_subparsers(dest="command")
    subparsers.add_parser("pdobjects", help="list supported Pure Data objects")
    subparser_hvhash = subparsers.add_parser("hvhash", help="print the heavy hash of the input string")
    subparser_hvhash.add_argument("string")
    subparser_meta = subparsers.add_parser("metaschema", help="show JSON Schema of the Meta() model")
    subparser_meta.add_argument("-g", "--generator", help="Choose supported generator: DPF, Daisy")

    parsed_args = parser.parse_args(args=None if sys.argv[1:] else ['--help'])
    args = vars(parsed_args)

    command = args.pop("command")
    if command == "pdobjects":
        obj_list = PdParser.get_supported_objects()
        obj_list.sort()

        obj_dict = {
            "Message Objects": [k for k in obj_list if '~' not in k],
            "Signal Objects": [k for k in obj_list if '~' in k]
        }
        print(json.dumps(obj_dict, indent=4))
    elif command == "hvhash":
        print("0x{0:X}".format(HeavyLangObject.get_hash(args['string'])))
    elif command == "metaschema":
        generator = args.get('generator')
        if generator is not None:
            model = gens[generator.lower()]()
        else:
            model = Meta
        print(json.dumps(model.model_json_schema(), indent=4))
    else:
        pass


if __name__ == "__main__":
    main()



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional, List, Dict

from .NotificationEnum import NotificationEnum
from .PdObject import PdObject


class PdTableObject(PdObject):

    def __init__(
        self,
        obj_type: str,
        obj_args: Optional[List] = None,
        pos_x: int = 0,
        pos_y: int = 0
    ) -> None:
        assert obj_type == "table"
        super().__init__(obj_type, obj_args, pos_x, pos_y)

        self.__table_name = ""
        self.__size = 0
        self.__extern = False

        try:
            self.__table_name = self.obj_args[0]
        except Exception:
            self.add_error(
                "Missing \"name\" argument for table",
                NotificationEnum.ERROR_MISSING_REQUIRED_ARGUMENT)

        try:
            # optional arguments
            try:
                self.__size = int(self.obj_args[1])
                self.__extern = (self.obj_args[2] == "@hv_table")
            except Exception:
                pass
        except Exception:
            pass

    def to_hv(self) -> Dict:
        return {
            "type": "table",
            "args": {
                "name": self.__table_name,
                "size": self.__size,
                "values": [],
                "extern": self.__extern
            },
            "properties": {
                "x": self.pos_x,
                "y": self.pos_y
            },
            "annotations": {
                "scope": "public"
            }
        }



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023-2024 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Callable, Dict, List

from .HeavyObject import HeavyObject

from hvcc.types.IR import IRObjectdict, IROnMessage


class ControlPrint(HeavyObject):
    """Prints the first value in a message to the console"""

    preamble = "cPrint"

    @classmethod
    def get_C_header_set(cls) -> set:
        return {"HvControlPrint.h"}

    @classmethod
    def get_C_file_set(cls) -> set:
        return {"HvControlPrint.h", "HvControlPrint.c"}

    @classmethod
    def get_C_init(cls, obj_type: str, obj_id: str, args: Dict) -> List[str]:
        return []

    @classmethod
    def get_C_free(cls, obj_type: str, obj_id: str, args: Dict) -> List[str]:
        return []

    @classmethod
    def get_C_onMessage(cls, obj_type: str, obj_id: str, inlet_index: int, args: Dict) -> List[str]:
        return [f"cPrint_onMessage(_c, m, \"{args['label']}\");"]

    @classmethod
    def get_C_decl(cls, obj_type: str, obj_id: str, args: Dict) -> List[str]:
        return []

    @classmethod
    def get_C_impl(
        cls,
        obj_type: str,
        obj_id: str,
        on_message_list: List[List[IROnMessage]],
        get_obj_class: Callable,
        objects: Dict[str, IRObjectdict]
    ) -> List[str]:
        return []
